/*  SciGraphica - Scientific graphics and data manipulation
 *  Copyright (C) 2000 Adrian E. Feiguin <feiguin@ifir.edu.ar>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtkextra/gtkextra.h>

#include "main.h"

/* TODO
 * dataset function 
 */

static void save_plot			(SGplotfile *file);
static void save_layer			(SGplotfile *file, SGlayer *layer);
static void save_dataset		(SGplotfile *file, SGdataset *dataset);
static void save_axis			(SGplotfile *file, GtkPlotAxis *axis);
static void set_line_style		(SGplotfile *file, GtkPlotLine line);
static void set_symbol_style		(SGplotfile *file, 
					 GtkPlotSymbol symbol);
static void put_text			(SGplotfile *file, GtkPlotText *text);
static void put_child			(SGplotfile *file, GtkPlotCanvasChild *child);

void
sg_plot_file_export_xml(SGplot *plot, gchar *filename, FILE* opened)
{
  SGplotfile *file;

  file = sg_plot_file_new(plot, filename);

  if(!opened){
    if ((file->file = fopen(file->filename, "w")) == NULL){
         g_warning("ERROR: Cannot open file: %s", filename);
         g_free(file);
         return;
    }
    fprintf(file->file,"<?xml version=\"1.0\"?>\n");
  } else {
    file->file = opened;
  }

  file->save_plot = save_plot;
  file->save_layer = save_layer;
  file->save_dataset = save_dataset;
  file->save_axis = save_axis;
  file->set_line_style = set_line_style;
  file->set_symbol_style = set_symbol_style;
  file->put_text = put_text;
  file->put_child = put_child;

  sg_plot_file_export(file);

  if(!opened){
    fclose(file->file);
  }
}

static void
save_plot(SGplotfile *file)
{
  SGplot *plot = file->plot;
  GtkPlotCanvas *canvas;
  GList *list;

  canvas = GTK_PLOT_CANVAS(plot->real_canvas);

  fprintf(file->file,"<sgp:Plot xmlns:sgp=\"http://scigraphica.sourceforge.net\">\n");

  fprintf(file->file, "  <sgp:Summary>\n");
  fprintf(file->file, "    <sgp:Item>\n");
  fprintf(file->file, "      <sgp:name>application</sgp:name>\n");
  fprintf(file->file, "      <sgp:val-string>scigraphica</sgp:val-string>\n");
  fprintf(file->file, "    </sgp:Item>\n");
  fprintf(file->file, "    <sgp:Item>\n");
  fprintf(file->file, "      <sgp:name>author</sgp:name>\n");
  fprintf(file->file, "      <sgp:val-string>%s</sgp:val-string>\n", g_get_real_name());
  fprintf(file->file, "    </sgp:Item>\n");
  fprintf(file->file, "  </sgp:Summary>\n");
  fprintf(file->file, "  <sgp:Geometry Width=\"%d\" Height=\"%d\"/>\n",
          plot->width, plot->height);
  fprintf(file->file, "  <sgp:Page Orientation=\"%d\" Size=\"%d\" Width=\"%d\" Height=\"%d\" Scale=\"%f\"/>\n", plot->orientation, plot->page_size, plot->page_width, plot->page_height, plot->scale);

  fprintf(file->file, "     <sgp:Background R=\"%d\" G=\"%d\" B=\"%d\"/>\n", canvas->background.red, canvas->background.green, canvas->background.blue); 
  fprintf(file->file, "  <sgp:Name>%s</sgp:Name>\n", plot->name);

  list = canvas->childs;
  while(list)
   {
      GtkPlotCanvasChild *child;
      child = (GtkPlotCanvasChild *)list->data;
      file->put_child(file, child);
     
      list = list->next;
   }

  list = plot->layers;
  while(list)
   {
      file->save_layer(file, (SGlayer *)list->data);
     
      list = list->next;
   }

  fprintf(file->file,"</sgp:Plot>\n");
}

static void
save_layer(SGplotfile *file, SGlayer *layer)
{
  GtkPlot *plot = GTK_PLOT(layer->real_plot);
  GList *list;

/*********************************************************************
 * LAYER PLOT 
 *********************************************************************/

  fprintf(file->file, "  <sgp:Layer>\n");
  fprintf(file->file, "    <sgp:Geometry X=\"%f\" Y=\"%f\" Width=\"%f\" Height=\"%f\" Magnification=\"%f\" Xscale=\"%d\" Yscale=\"%d\"/>\n", plot->x, plot->y, plot->width, plot->height, plot->magnification, plot->xscale, plot->yscale);
  fprintf(file->file, "    <sgp:Range Xmin=\"%f\" Xmax=\"%f\" Ymin=\"%f\" Ymax=\"%f\"/>\n", plot->xmin, plot->xmax, plot->ymin, plot->ymax);
  fprintf(file->file, "    <sgp:Background R=\"%d\" G=\"%d\" B=\"%d\" Transparent=\"%d\"/>\n", plot->background.red, plot->background.green, plot->background.blue, GTK_PLOT_TRANSPARENT(plot)); 

  fprintf(file->file, "    <sgp:Vmajor Visible=\"%d\">\n", plot->show_vmajor); 
  file->set_line_style(file, plot->major_vgrid);
  fprintf(file->file, "    </sgp:Vmajor>\n"); 

  fprintf(file->file, "    <sgp:Vminor Visible=\"%d\">\n", plot->show_vminor); 
  file->set_line_style(file, plot->minor_vgrid);
  fprintf(file->file, "    </sgp:Vminor>\n"); 

  fprintf(file->file, "    <sgp:Hmajor Visible=\"%d\">\n", plot->show_hmajor); 
  file->set_line_style(file, plot->major_hgrid);
  fprintf(file->file, "    </sgp:Hmajor>\n"); 

  fprintf(file->file, "    <sgp:Hminor Visible=\"%d\">\n", plot->show_hminor); 
  file->set_line_style(file, plot->minor_hgrid);
  fprintf(file->file, "    </sgp:Hminor>\n"); 

  fprintf(file->file, "    <sgp:X0 Visible=\"%d\">\n", plot->show_x0); 
  file->set_line_style(file, plot->x0_line);
  fprintf(file->file, "    </sgp:X0>\n"); 

  fprintf(file->file, "    <sgp:Y0 Visible=\"%d\">\n", plot->show_y0); 
  file->set_line_style(file, plot->y0_line);
  fprintf(file->file, "    </sgp:Y0>\n"); 

/*********************************************************************
 * AXES 
 *********************************************************************/

  fprintf(file->file, "    <sgp:Axis Position=\"Bottom\" Align=\"%f\" Visible=\"%d\" Orientation=\"%d\" Scale=\"%d\">\n", plot->bottom_align, plot->bottom.is_visible, plot->bottom.orientation, plot->bottom.scale_type); 
  file->save_axis(file, &plot->bottom);
  fprintf(file->file, "    </sgp:Axis>\n"); 
  fprintf(file->file, "    <sgp:Axis Position=\"Top\" Align=\"%f\" Visible=\"%d\" Orientation=\"%d\" Scale=\"%d\">\n", plot->top_align, plot->top.is_visible, plot->top.orientation, plot->top.scale_type); 
  file->save_axis(file, &plot->top);
  fprintf(file->file, "    </sgp:Axis>\n"); 
  fprintf(file->file, "    <sgp:Axis Position=\"Left\" Align=\"%f\" Visible=\"%d\" Orientation=\"%d\" Scale=\"%d\">\n", plot->left_align, plot->left.is_visible, plot->left.orientation, plot->left.scale_type); 
  file->save_axis(file, &plot->left);
  fprintf(file->file, "    </sgp:Axis>\n"); 
  fprintf(file->file, "    <sgp:Axis Position=\"Right\" Align=\"%f\" Visible=\"%d\" Orientation=\"%d\" Scale=\"%d\">\n", plot->right_align, plot->right.is_visible, plot->right.orientation, plot->right.scale_type); 
  file->save_axis(file, &plot->right);
  fprintf(file->file, "    </sgp:Axis>\n"); 

/*********************************************************************
 * TICKS 
 *********************************************************************/

  fprintf(file->file, "    <sgp:Xticks Step=\"%f\" Nminor=\"%d\" Limits=\"%d\" Begin=\"%f\" End=\"%f\">\n", plot->xticks.step, plot->xticks.nminor, plot->xticks.set_limits, plot->xticks.begin, plot->xticks.end); 

  fprintf(file->file, "    </sgp:Xticks>\n"); 

  fprintf(file->file, "    <sgp:Yticks Step=\"%f\" Nminor=\"%d\" Limits=\"%d\" Begin=\"%f\" End=\"%f\">\n", plot->yticks.step, plot->yticks.nminor, plot->yticks.set_limits, plot->yticks.begin, plot->yticks.end); 

  fprintf(file->file, "    </sgp:Yticks>\n"); 

/*********************************************************************
 * LEGENDS
 *********************************************************************/

  fprintf(file->file, "    <sgp:Legends X=\"%f\" Y=\"%f\" Width=\"%d\" Height=\"%d\" Visible=\"%d\">\n", plot->legends_x, plot->legends_y, plot->legends_width, plot->legends_height, plot->show_legends); 
  fprintf(file->file, "      <sgp:Border Style=\"%d\" Line=\"%d\" Shadow=\"%d\"/>\n", plot->legends_border, plot->legends_border_width, plot->legends_shadow_width); 
 
  file->put_text(file, &plot->legends_attr);
  fprintf(file->file, "    </sgp:Legends>\n"); 

/*********************************************************************
 * DATASETS 
 *********************************************************************/

  list = layer->datasets;
  while(list)
   {
      file->save_dataset(file, (SGdataset *)list->data);
     
      list = list->next;
   }

  fprintf(file->file, "  </sgp:Layer>\n");
}


static void
save_dataset(SGplotfile *file, SGdataset *dataset)
{
  GtkPlotData *data = dataset->real_data;
  gint n = 0;


/*********************************************************************
 * DATASET 
 *********************************************************************/
  if(data->is_function)
    fprintf(file->file, "    <sgp:Dataset Visible=\"%d\" Labels=\"%d\" Connector=\"%d\" Name=\"%s\" Function=\"%s\">\n", data->is_visible, data->show_labels, data->line_connector, data->name, dataset->exp); 

  else
    fprintf(file->file, "    <sgp:Dataset Visible=\"%d\" Labels=\"%d\" Connector=\"%d\" Name=\"%s\">\n", data->is_visible, data->show_labels, data->line_connector, data->name); 

  
  if(data->legend)
    fprintf(file->file, "      <sgp:Legend Visible=\"%d\">%s</sgp:Legend>\n", data->show_legend, data->legend); 
  else
    fprintf(file->file, "      <sgp:Legend Visible=\"%d\"/>\n", data->show_legend); 

  if(dataset->worksheet)
    fprintf(file->file, "      <sgp:Datasheet Name=\"%s\" X=\"%d\" Y=\"%d\" DX=\"%d\" DY=\"%d\" Labels=\"%d\"/>\n", dataset->worksheet->name, dataset->x, dataset->y, dataset->dx, dataset->dy, dataset->l); 


  file->set_symbol_style(file, data->symbol);
  file->set_line_style(file, data->line);
   
  fprintf(file->file, "      <sgp:Xline>\n"); 
  file->set_line_style(file, data->x_line);
  fprintf(file->file, "      </sgp:Xline>\n"); 
     
  fprintf(file->file, "      <sgp:Yline>\n"); 
  file->set_line_style(file, data->y_line);
  fprintf(file->file, "      </sgp:Yline>\n"); 
     

/*********************************************************************
 * POINTS 
 *********************************************************************/
  fprintf(file->file, "      <sgp:Points No=\"%d\">\n", data->num_points); 

  if(data->x && data->y){
    gdouble dx = 0.;
    gdouble dy = 0.;

    if(data->dx) dx = data->dx[n];
    if(data->dy) dy = data->dy[n];

    for(n = 0; n < dataset->real_data->num_points; n++){
      fprintf(file->file, "        <sgp:Point X=\"%f\" Y=\"%f\" DX=\"%f\" DY=\"%f\"/>\n", data->x[n], data->y[n], dx, dy); 
    }
  }

  fprintf(file->file, "      </sgp:Points>\n"); 

/*********************************************************************
 * X ERRBARS 
 *********************************************************************/
  fprintf(file->file, "      <sgp:Xerrbars Visible=\"%d\" Width=\"%d\" Caps=\"%d\"/>\n", data->show_xerrbars, data->xerrbar_width, data->xerrbar_caps); 
/*********************************************************************
 * Y ERRBARS 
 *********************************************************************/
  fprintf(file->file, "      <sgp:Yerrbars Visible=\"%d\" Width=\"%d\" Caps=\"%d\"/>\n", data->show_yerrbars, data->yerrbar_width, data->yerrbar_caps); 
/*********************************************************************
 * LABELS 
 *********************************************************************/
  fprintf(file->file, "      <sgp:Datalabels Visible=\"%d\" Offset=\"%d\">\n", data->show_labels, data->labels_offset); 
  
  file->put_text(file, &data->labels_attr);

  if(data->labels){
    for(n = 0; n < dataset->real_data->num_points; n++){
      if(data->labels[n])
        fprintf(file->file, "        <sgp:Label Point=\"%d\" Text=\"%s\"/>\n", n, data->labels[n]); 
    }
  }

  fprintf(file->file, "      </sgp:Datalabels>\n"); 
/*********************************************************************/
  fprintf(file->file, "    </sgp:Dataset>\n"); 

}

static void
save_axis(SGplotfile *file, GtkPlotAxis *axis)
{
  gint n = 0;

/*********************************************************************
 * AXIS 
 *********************************************************************/

  fprintf(file->file, "      <sgp:Ticks Major=\"%d\" Minor=\"%d\" Length=\"%d\" Width=\"%d\"/>\n", axis->major_mask, axis->minor_mask, axis->ticks_length, axis->ticks_width); 
  file->set_line_style(file, axis->line);
  fprintf(file->file, "      <sgp:Title Visible=\"%d\">\n", axis->title_visible); 

  file->put_text(file, &axis->title);
  fprintf(file->file, "      </sgp:Title>\n"); 

  fprintf(file->file, "      <sgp:Labels Precision=\"%d\" Style=\"%d\" Mask=\"%d\" Offset=\"%d\">\n", axis->label_precision, axis->label_style, axis->label_mask, axis->labels_offset); 

  file->put_text(file, &axis->labels_attr);

  if(axis->custom_labels){
    fprintf(file->file, "        <sgp:Ticklabels No=\"%d\">\n", axis->nlabels); 
    if(axis->tick_labels){
      for(n = 0; n < axis->nlabels; n++){
        fprintf(file->file, "          <sgp:Ticklabel Text=\"%s\"/>\n", axis->tick_labels[n]); 
      }
    }  
    fprintf(file->file, "        </sgp:Ticklabels>\n"); 
  }

  fprintf(file->file, "      </sgp:Labels>\n"); 
}


/*********************************************************************
 * LINE STYLE 
 *********************************************************************/
static void
set_line_style(SGplotfile *file, GtkPlotLine line)
{
  fprintf(file->file, "        <sgp:Line Style=\"%d\" Width=\"%d\" R=\"%d\" G=\"%d\" B=\"%d\"/>\n", line.line_style, line.line_width, line.color.red, line.color.green, line.color.blue); 

}

/*********************************************************************
 * SYMBOL STYLE 
 *********************************************************************/
static void
set_symbol_style(SGplotfile *file, GtkPlotSymbol symbol)
{
  fprintf(file->file, "        <sgp:Symbol Type=\"%d\" Style=\"%d\" Size=\"%d\" Width=\"%d\" R=\"%d\" G=\"%d\" B=\"%d\"/>\n", symbol.symbol_type, symbol.symbol_style, symbol.size, symbol.line_width, symbol.color.red, symbol.color.green, symbol.color.blue); 
}

/*********************************************************************
 * TEXT 
 *********************************************************************/
static void
put_text(SGplotfile *file, GtkPlotText *text)
{
  fprintf(file->file, "        <sgp:Text X=\"%f\" Y=\"%f\" Angle=\"%d\" Justification=\"%d\" Transparent=\"%d\">\n", text->x, text->y, text->angle, text->justification, text->transparent); 
  fprintf(file->file, "          <sgp:Border Style=\"%d\" Width=\"%d\" Shadow=\"%d\"/>\n", text->border, text->border_width, text->shadow_width); 
  fprintf(file->file, "          <sgp:Foreground R=\"%d\" G=\"%d\" B=\"%d\"/>\n", text->fg.red, text->fg.green, text->fg.blue); 
  fprintf(file->file, "          <sgp:Background R=\"%d\" G=\"%d\" B=\"%d\"/>\n", text->bg.red, text->bg.green, text->bg.blue); 
  fprintf(file->file, "          <sgp:Font Name=\"%s\" Height=\"%d\"/>\n", text->font, text->height); 

  if(text->text && strlen(text->text) > 0){
           gchar *xml_text = NULL;
           xml_text = xml_get_string(text->text);
           fprintf(file->file, "<sgp:String>%s</sgp:String>\n", xml_text); 
           g_free(xml_text);
  }

  fprintf(file->file, "        </sgp:Text>\n"); 
}


static void
put_child(SGplotfile *file, GtkPlotCanvasChild *child)
{
  GtkPlotCanvasLine *line = NULL;
  GtkPlotCanvasRectangle *rectangle = NULL;
  GtkPlotCanvasEllipse *ellipse = NULL;

  fprintf(file->file, "        <sgp:Child X1=\"%f\" Y1=\"%f\" X2=\"%f\" Y2=\"%f\" Type=\"%d\" Flags=\"%d\">\n", child->rx1, child->ry1, child->rx2, child->ry2, child->type, child->flags); 

  switch(child->type){
   case GTK_PLOT_CANVAS_TEXT:
       file->put_text(file, (GtkPlotText *)child->data);
       break;
   case GTK_PLOT_CANVAS_LINE:
       line = (GtkPlotCanvasLine *)child->data;
       fprintf(file->file, "          <sgp:Arrow Mask=\"%d\" Length=\"%d\">\n",
               line->arrow_mask, line->arrow_length); 
       file->set_line_style(file, line->line);
       fprintf(file->file, "          </sgp:Arrow>\n"); 
       break;
   case GTK_PLOT_CANVAS_RECTANGLE:
       rectangle = (GtkPlotCanvasRectangle *)child->data;
       fprintf(file->file, "          <sgp:Rectangle Fill=\"%d\" Border=\"%d\" Shadow=\"%d\" R=\"%d\" G=\"%d\" B=\"%d\">\n", rectangle->filled, rectangle->border, rectangle->shadow_width, rectangle->bg.red, rectangle->bg.green, rectangle->bg.blue); 
       file->set_line_style(file, rectangle->line);
       fprintf(file->file, "          </sgp:Rectangle>\n"); 
       break;
   case GTK_PLOT_CANVAS_ELLIPSE:
       ellipse = (GtkPlotCanvasEllipse *)child->data;
       fprintf(file->file, "          <sgp:Ellipse Fill=\"%d\" R=\"%d\" G=\"%d\" B=\"%d\">\n", ellipse->filled, ellipse->bg.red, ellipse->bg.green, ellipse->bg.blue); 
       file->set_line_style(file, ellipse->line);
       fprintf(file->file, "          </sgp:Ellipse>\n"); 
       break;
   default:
  }

  fprintf(file->file, "        </sgp:Child>\n"); 

}
