/*  SciGraphica - Scientific graphics and data manipulation
 *  Copyright (C) 2000 Adrian E. Feiguin <feiguin@ifir.edu.ar>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>
#include <string.h>
#include <gtkextra/gtkextra.h>
#include "main.h"

static void remove_worksheet			(SGworksheet *worksheet);
static void remove_plot				(SGplot *plot);
static gint rename_worksheet			(GtkIconList *iconlist, 
						 GtkIconListItem *item);
static gint rename_plot				(GtkIconList *iconlist, 
						 GtkIconListItem *item);


void
sg_project_new()
{
 sg_project_init();

 gtk_signal_connect(GTK_OBJECT(iconlist2), "deactivate_icon",
                    (GtkSignalFunc)rename_plot, iconlist2);

 gtk_signal_connect(GTK_OBJECT(iconlist1), "deactivate_icon",
                    (GtkSignalFunc)rename_worksheet, iconlist1);
}

void
sg_project_init()
{
 last_function = num_functions = 0;
 last_plot = num_plots = 0;
 last_worksheet = num_worksheets = 0;
 active_icon = NULL;
 active_worksheet = NULL;
 active_plot = NULL;
 active_layer = NULL;
 active_dataset = NULL;

 sg_clipboard_init();
 sg_dialogs_init();
 plots = NULL;
 worksheets = NULL;
 functions = NULL;

 if(last_project_filename){
    g_free(last_project_filename);
    last_project_filename = g_strdup("untitled.sg");
 }

}


SGworksheet *
sg_project_new_worksheet()
{
 SGworksheet *worksheet;
 gchar *name;
 gchar label[20];

 num_worksheets++;
 last_worksheet++;
 sprintf(label,"Data%d",last_worksheet);
 name = g_strdup(label);

 worksheet = sg_worksheet_new(GTK_ICON_LIST(iconlist1), name, 20, 5);
 gtk_notebook_set_page(GTK_NOTEBOOK(gui_notebook), 0);

 gtk_signal_connect(GTK_OBJECT(worksheet->sheet), "button_press_event",
                    (GtkSignalFunc)sg_worksheet_menu_show, NULL);
 active_worksheet = worksheet;

 return worksheet;
}

SGplot *
sg_project_new_plot()
{
 SGplot *plot;
 gchar *name;
 gchar label[20];

 num_plots++;
 last_plot++;
 sprintf(label,"Plot%d",last_plot);
 name = g_strdup(label);

 plot = sg_plot_new_with_layer(GTK_ICON_LIST(iconlist2), name);
 gtk_notebook_set_page(GTK_NOTEBOOK(gui_notebook), 1);

 active_plot = plot;

 return plot;
}

void
sg_project_add_function(SGdataset *function)
{
  functions = g_list_append(functions, function);
  num_functions++;
  last_function++;
}


void
sg_project_remove_worksheet(SGworksheet *worksheet)
{
  remove_worksheet(worksheet);    
}

void
sg_project_remove_plot(SGplot *plot)
{
  remove_plot(plot);    
}

void
sg_project_close()
{
  GList *list;

#ifndef WITH_GNOME
  gtk_signal_disconnect_by_data(GTK_OBJECT(iconlist1), iconlist1);
  gtk_signal_disconnect_by_data(GTK_OBJECT(iconlist2), iconlist2);
#endif
  gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist1), NULL);
  gtk_icon_list_set_active_icon(GTK_ICON_LIST(iconlist2), NULL);

  list = plots;
  while(list){
    remove_plot((SGplot *)list->data);
    list = plots;
  }

  list = worksheets;
  while(list){
    remove_worksheet((SGworksheet *)list->data);
    list = worksheets;
  }

  list = functions;
  while(list){
    functions = g_list_remove_link(functions, list);
    g_list_free_1(list);
    list = functions;
  }

  sg_clipboard_clear();
  sg_dialogs_destroy();
}

static void
remove_worksheet(SGworksheet *worksheet)
{
  GList *list;

  if(worksheet == NULL) return;

  list = g_list_find(worksheets, worksheet);

  if(!list) return;

  sg_worksheet_remove(worksheet);

  worksheets = g_list_remove_link(worksheets, list);
  g_list_free_1(list);

  num_worksheets--;

  if(num_worksheets == 0) worksheets = NULL;
}

static void
remove_plot(SGplot *plot)
{
  GList *list;

  if(plot == NULL) return;

  list = g_list_find(plots, plot);

  if(!list) return;

  sg_plot_remove(plot);

  plots = g_list_remove_link(plots, list);
  g_list_free_1(list);

  num_plots--;

  if(num_plots == 0) plots = NULL;
}

static gint
rename_worksheet(GtkIconList *iconlist, GtkIconListItem *item)
{
  GList *list;
  gint veto;
  gchar *name;
  SGworksheet *worksheet, *data;
  gint n;

  worksheet = (SGworksheet *)gtk_icon_list_get_link(item);
  name = gtk_entry_get_text(GTK_ENTRY(item->entry));

  if(name == NULL || strlen(name) == 0){
    sg_message_dialog("The worksheet must have a name", 0);
    return FALSE;
  }

  for(n = 0; n < strlen(name); n++){
    if(name[n] < 'a' || name[n] > 'z')
     if(name[n] < 'A' || name[n] > 'Z')
      if(name[n] < '0' || name[n] > '9')
        if(name[n] != '.' && name[n] != '_' && name[n] != '-' && name[n] != '+' && name[n] != '='){
          sg_message_dialog("The name contains invalid characters", 0);
          return FALSE;
      }
  }

  list = worksheets;
  while(list){
   data = (SGworksheet *)list->data;

   if(data!= worksheet && strcmp(data->name, name) == 0){
      sg_message_dialog("Another worksheet has the same name", 0);
      return FALSE;
   }

   list = list->next;
  }


  veto = sg_worksheet_rename(worksheet, name);
  return veto;

}

static gint
rename_plot(GtkIconList *iconlist, GtkIconListItem *item)
{
  GList *list;
  gint veto;
  gchar *name;
  SGplot *plot, *data;
  gint n;

  plot = (SGplot *)gtk_icon_list_get_link(item);
  name = gtk_entry_get_text(GTK_ENTRY(item->entry));

  if(name == NULL || strlen(name) == 0){
    sg_message_dialog("The plot must have a name", 0);
    return FALSE;
  }

  for(n = 0; n < strlen(name); n++){
    if(name[n] < 'a' || name[n] > 'z')
     if(name[n] < 'A' || name[n] > 'Z')
      if(name[n] < '0' || name[n] > '9')
        if(name[n] != '.' && name[n] != '_' && name[n] != '-' && name[n] != '+' && name[n] != '='){
          sg_message_dialog("The name contains invalid characters", 0);
          return FALSE;
      }
  }

  list = plots;
  while(list){
   data = (SGplot *)list->data;

   if(data!= plot && strcmp(data->name, name) == 0){
      sg_message_dialog("Another plot has the same name", 0);
      return FALSE;
   }

   list = list->next;
  }


  veto = sg_plot_rename(plot, name);
  return veto;

}

void
sg_project_open_worksheet(GtkWidget *widget, 
                          GtkIconListItem *item, GdkEvent *event,
                          gpointer data)
{

  GtkIconList *iconlist;
  SGworksheet *worksheet;
  GdkModifierType mods;
  iconlist = GTK_ICON_LIST(widget);

  gdk_window_get_pointer(widget->window, NULL, NULL, &mods);

  worksheet = (SGworksheet *)gtk_icon_list_get_link(item);

  if(!event) return;
  if((mods & GDK_BUTTON1_MASK) && event->type == GDK_2BUTTON_PRESS)
                sg_worksheet_open(worksheet);

  if(mods & GDK_BUTTON3_MASK)
                sg_worksheet_popup_show(item,(GdkEventButton *)event);

  active_worksheet = worksheet;

  return;

}

void
sg_project_open_plot(GtkWidget *widget, 
                     GtkIconListItem *item,
                     GdkEvent *event,
                     gpointer data)
{
  GtkIconList *iconlist;
  SGplot *plot;
  GdkModifierType mods;
  iconlist = GTK_ICON_LIST(widget);

  gdk_window_get_pointer(widget->window, NULL, NULL, &mods);

  plot = (SGplot *)gtk_icon_list_get_link(item);
  if(!event) return;
  if((mods & GDK_BUTTON1_MASK) && event->type == GDK_2BUTTON_PRESS){
                sg_plot_open(plot);
  }

  if(mods & GDK_BUTTON3_MASK)
                sg_plot_popup_show(item,(GdkEventButton *)event);

  active_plot = plot;

  return;

}
